/*
 * Copyright (c) 1996-1998 PFU Limited.
 *	author Kazuhisa Shirakami
 */
package teikade.gui.editor;

import java.awt.*;
import teikade.gui.*;
import java.util.Vector;
import java.util.Hashtable;
import java.util.BitSet;
import teikade.sys.SystemManager;

public class EditField extends TextField implements EditDataViewer {
    private EditData editData = new EditData(this);
    private BitSet keys;		// for JDK bugs on Win32
    // thanks for JDK bug patch on Win32 by Hideyo Namimoto
    private Win32TextComponentPatch win32patch;

    public EditField() {
	super();
	initialize();
    }

    public EditField(int cols) {
	super(cols);
	initialize();
    }

    public EditField(String text) {
	super(text);
	initialize();
    }

    public EditField(String text, int cols) {
	super(text, cols);
	initialize();
    }

    void initialize() {
	if (SystemManager.onWin32) {
	    keys = new BitSet(256);
	    // thanks for for JDK bug patch on Win32 by Hideyo Namimoto
	    if (SystemManager.needWin32TextComponentPatch) {
		win32patch = new Win32TextComponentPatch();
	    }
	}
    }

    public EditData getEditData() {
	return editData;
    }

    public boolean keyDown(Event event, int key) {
	if (key == '\n' || key == '\t') return false;
	if (editData != null) {
	    EditKeyHandler keyHandler = EditProperties.getKeyHandler(isEditable());
	    if (keyHandler != null) {
		editData.update();
		editData.pushStatus();
		if (keys != null && event.key > 0 && event.key < 256) {
		    keys.set(event.key);
		}
		if (win32patch != null) win32patch.patchEvent(event);
		return keyHandler.handleKey(editData, event);
	    }
	}
	return super.keyDown(event, key);
    }

    public boolean keyUp(Event event, int key) {
	if (key == '\n' || key == '\t') return false;
	if (keys != null && editData != null
	&&  event.key > 0 && event.key < 256
	&&  EditProperties.getKeyHandler(isEditable()) != null) {
	    if ((event.modifiers & Event.CTRL_MASK) != 0
	    &&  !keys.get(event.key)) {
		keyDown(event, key);
	    }
	    keys.clear(event.key);
	    return true;
	}
	return super.keyUp(event, key);
    }

    // thanks for JDK bug patch on Win32 by Hideyo Namimoto
    public String getText() {
	String text = super.getText();
	if (win32patch != null) win32patch.setText(text);
	return text;
    }

    public void setText(String text) {
	super.setText(text);
	// thanks for JDK bug patch on Win32 by Hideyo Namimoto
	if (win32patch != null) win32patch.setText(text);
	editData.reset();
    }

    public void insertText(String text, int pos) {
	int current = getSelectionStart();
	String org = super.getText();
	setText(org.substring(0, pos) + text + org.substring(pos));
	// thanks for JDK bug patch on Win32 by Hideyo Namimoto
	if (win32patch != null) getText();
	if (current >= pos) current += text.length();
	select(current, current);
    }

    public void replaceText(String text, int start, int end) {
	int current = getSelectionStart();
	String org = super.getText();
	setText(org.substring(0, start) + text + org.substring(end));
	// thanks for JDK bug patch on Win32 by Hideyo Namimoto
	if (win32patch != null) getText();
	if (current > start) {
	    if (current < end) {
		current = start;
	    } else {
		current += text.length() - (end - start);
	    }
	}
	select(current, current);
    }

    // thanks for JDK bug patch on Win32 by Hideyo Namimoto
    public void select(int start, int end) {
	if (win32patch != null) {
	    if (start == end) {
		int n = win32patch.adjustPoint(start);
		start += n;
		end   += n;
	    } else {
		start += win32patch.adjustRange(start);
		end   += win32patch.adjustRange(end);
	    }
	}
	super.select(start, end);
    }

    // thanks for JDK bug patch on Win32 by Hideyo Namimoto
    public int getSelectionStart() {
	int start = super.getSelectionStart();
	if (win32patch != null) start -= win32patch.adjustForGetSelection(start);
	return start;
    }

    // thanks for JDK bug patch on Win32 by Hideyo Namimoto
    public int getSelectionEnd()  {
	int end = super.getSelectionEnd();
	if (win32patch != null) end -= win32patch.adjustForGetSelection(end);
	return end;
    }

    public int displayLines() {
	return size().height / getFont().getSize();
    }
}
