package teikade.net;

import java.io.*;

public abstract class Packet {
    private String className;
    private StreamTokenizer tokenizer;
    private int arguments;
    private boolean noMoreArguments;

    public static Packet from(String string, String packageName) throws IllegalArgumentException {
	return from(new StringBufferInputStream(string), packageName); 
    }

    public static Packet from(InputStream stream, String packageName) throws IllegalArgumentException {
	StreamTokenizer tokenizer = new StreamTokenizer(stream);
	int type = nextToken(tokenizer);
	if (type == StreamTokenizer.TT_EOF) return null;
	if (type != StreamTokenizer.TT_WORD) {
	    throw new IllegalArgumentException("class name expected");
	}
	String className = packageName + tokenizer.sval;
	try {
	    type = nextToken(tokenizer);
	    if (type != '(') {
		throw new IllegalArgumentException("'(' expected");
	    }
	    Packet packet = (Packet)Class.forName(className).newInstance();
	    packet.init(tokenizer);
	    if (tokenizer.ttype != ')') {
		throw new IllegalArgumentException("')' expected");
	    }
	    return packet;
	} catch (ClassNotFoundException e) {
	    throw new IllegalArgumentException("Unrecognized packet class: " + className);
	} catch (InstantiationException e) {
	    throw new IllegalArgumentException("Instantiation failed for packet: " + className);
	} catch (IllegalAccessException e) {
	    throw new IllegalArgumentException("Illegal access for packet: " + className);
	}
    }

    private final void init(StreamTokenizer tokenizer) throws IllegalArgumentException {
	this.tokenizer = tokenizer;
	arguments = 0;
	noMoreArguments = false;
	init();
	if (arguments == 0) checkDelimiter();
    }

    // This method should be overridden
    public void init() throws IllegalArgumentException {
    }

    public abstract void execute(Object target);

    protected int getIntArgument() throws IllegalArgumentException {
	if (nextArgument() != StreamTokenizer.TT_NUMBER) {
	    throw argumentError("not integer");
	}
	int val = (int)tokenizer.nval;
	checkDelimiter();
	return val;
    }

    protected String getStringArgument() throws IllegalArgumentException {
	if (nextArgument() != '"') {
	    throw argumentError("not string");
	}
	String val = tokenizer.sval;
	checkDelimiter();
	return val;
    }

    protected String toStreamString(String arg) {
	return "\"" + arg + "\"";
    }

    private int nextArgument() throws IllegalArgumentException {
	arguments++;
	if (!noMoreArguments) {
	    int type = nextToken(tokenizer);
	    if (type != StreamTokenizer.TT_EOF) {
		return type;
	    }
	}
	throw argumentError("expected");
    }

    private IllegalArgumentException argumentError(String message) {
	message = argumentName(arguments) + " argument: " + message;
	return new IllegalArgumentException(message);
    }

    protected String argumentName(int n) {
	switch (n) {
	case  1: return "1st";
	case  2: return "2nd";
	case  3: return "3rd";
	default: return n + "th";
	}
    }

    private void checkDelimiter() throws IllegalArgumentException {
	int type = nextToken(tokenizer);
	switch (type) {
	case ')':
	    noMoreArguments = true;
	    return;
	case ',':
	    return;
	default:
	    throw argumentError("illegal");
	}
    }

    private static int nextToken(StreamTokenizer tokenizer) throws IllegalArgumentException {
	try {
	    return tokenizer.nextToken();
	} catch (IOException e) {
	    throw new IllegalArgumentException("I/O error");
	}
    }

    public String toString() {
	String name = getClass().getName();
	return name.substring(name.lastIndexOf('.') + 1)
		+ "(" + argumentsToString() + ")";
    }

    // This method should be overridden
    public String argumentsToString() {
	return "";
    }

    public static void main(String args[]) {
	Packet p = Packet.from("SetBreakPoint(\"TestClass\", 128)", "");
	System.out.println(p.toString());
    }
}
