/*
 * Copyright (c) 1995-1998 PFU Limited.
 *	author Osamu Satoh
 */
package teikade.sys;

import teikade.lang.*;
import teikade.util.StringSorter;
import teikade.util.StringSourceCode;
import teikade.util.CodeConverter;
import java.io.*;
import java.util.Hashtable;
import java.util.Enumeration;
import java.util.Vector;
import java.util.Date;

/**
 * package management class
 * children are instances of SourceFileManager
 */
public class PackageManager extends PathedManager {
    /**
     * all packages managed by the system
     * each package must have unique name
     */
    private static Hashtable packages = new Hashtable();

    /**
     * default header file for this project
     */
    private TeikadeFile header = null;

    /**
     * separator character of package name
     */
    public static final char SEPARATOR = '.';

    /**
     * replacement character for package separator character exists
     * in package path
     */
    private static final char SEPARATOR_REPLACEMENT = '~';

    /**
     * update sources/classes list file
     */
    public void updateSourcesFile() {
	SourceFileManager.clearLastSearchedCache();
	TeikadeFile srcsFile = sourcesFile();
	TeikadeFile backupFile = backupFile(srcsFile);
	TeikadeFile newFile = newFile(srcsFile);
	try {
	    //PrintStream ps = new PrintStream(new FileOutputStream(newFile));
	    PrintStream ps = new PrintStream(newFile.getOutputStream());
	    try {
		for (Enumeration c = children(); c.hasMoreElements(); ) {
		    SourceFileManager sfm = (SourceFileManager)c.nextElement();
		    sfm.writeListOn(ps);
		}
	    } finally {
		ps.close();
	    }
	    srcsFile.renameTo(backupFile);
	    newFile.renameTo(srcsFile);
	} catch(java.io.IOException e) {
	    SystemManager.exception("Could not write file: " + srcsFile.getPath(), e);
	}
    }

    public static String nameFor(String path) {
	return path.replace(SEPARATOR,SEPARATOR_REPLACEMENT).replace(File.separatorChar, SEPARATOR);
    }

    public static String pathFor(String name) {
	return name.replace(SEPARATOR, File.separatorChar).replace(SEPARATOR_REPLACEMENT,SEPARATOR);
    }

    public static String topName(Manager parent) {
	return "@" + parent.name();
    }

    public static PackageManager findPackage(String name) {
	if (name == null) {
	    return (PackageManager)packages.elements().nextElement();
	} else if (packages.containsKey(name)) {
	    return (PackageManager)packages.get(name);
	} else {
	    return null;
	}
    }

    public static PackageManager forName(PathedManager parent, String name, boolean readOnly) {
	if (name.equals(topName(parent))) {
	    return new TopDirPackageManager(parent);
	} else {
	    return new PackageManager(parent, name, pathFor(name));
	}
    }

    public static PackageManager forName(PathedManager parent, String name) {
	if (name.equals(topName(parent))) {
	    return new TopDirPackageManager(parent);
	} else {
	    return new PackageManager(parent, name, pathFor(name));
	}
    }

    /**
     * the constructor
     */
    PackageManager(Manager parent, String name, String path) {
	super(parent, name, path);
	packages.put(name, this);
    }

    /**
     * enumerate all packages managed by the system
     */
    public static Enumeration allPackages() {
	return packages.elements();
    }

    /**
     * get package CodePath
     */
    public CodePath getCodePath() {
	if (parent != null) {
	    CodePath cp = parent.getCodePath();
	    cp.packageManager(this);
	    return cp;
	} else {
	    return new CodePath(null, null, this, null, null);
	}
    }

    /**
     * whether it is editable or not
     */
    public boolean isEditable() {
	if (commentFile().exists()) {
	    return commentFile().canWrite();
	} else {
	    return directory().canWrite();
	}
    }

    /**
     * whether it has default header or not
     */
    public boolean hasDefaultHeader(boolean mustBeEditable) {
	if (headerFile().exists()) {
	    if (headerFile().canRead()) {
		return mustBeEditable ? headerFile().canWrite() : true;
	    }
	    SystemManager.information("Warning: The default header file for " + name +
					" exists but cannot read");
	}
	return false;
    }

    /**
     * set comment for this Package
     */
    public Manager setCode(String newComment) {
	CodeReader reader = new CodeReader(newComment);
	try {
	    int nest = reader.nextDescription();
	    if (nest > 0) {
		ClassSpec spec = new ClassSpec(reader.description());
		return createClass(spec, newComment);
	    }
	} catch (DescriptionException e) {
	    // It's not class description.
	}
	try {
	    TeikadeFile commentFile = commentFile();
	    TeikadeFile backupFile = backupFile(commentFile);
	    TeikadeFile newFile = newFile(commentFile);
	    CodeConverter.writeBytes(newFile, newComment);
	    commentFile.renameTo(backupFile);
	    newFile.renameTo(commentFile);
	    return this;
	} catch (java.io.IOException e) {
	    return null;
	}
    }

    /**
     * get package comment
     */
    public String getCodeInternal() throws java.io.IOException {
	TeikadeFile commentFile = commentFile();
	if (commentFile.exists()) {
	    return CodeConverter.toString(commentFile);
	} else {
	    return "<body>Package comment for JavaDoc</body>";
	}
    }

    /**
     * full name for this package
     */
    String fullName() {
	if (parent != null) {
	    return parent.name() + "." + name();
	} else {
	    return name();
	}
    }

    /**
     * get File object for package comment
     */
    TeikadeFile commentFile() {
	return TeikadeFile.create(directory(), "package.html");
    }

    /**
     * get File object for project default header
     */
    TeikadeFile headerFile() {
	if (header == null) {
	    header = TeikadeFile.create(directory(), ".header");
	}
	return header;
    }

    /**
     * get database file contains class list of this package
     */
    TeikadeFile sourcesFile() {
	return TeikadeFile.create(((ProjectManager)parent).profileDirectory(), fullName());
    }

    /**
     * synchronize all source files and class list database on sourcesFile
     */
    public void sync() {
	super.sync();
	updateSourcesFile();
    }

    /**
     * read sources/classes list file
     */
    void readSourcesFile() {
	/*
	if (child != null) {
	    disposeChildren();
	}
	*/
	TeikadeFile srcsfile = sourcesFile();
	try {
	    String line;
	    String pr = null;
	    String buffer = CodeConverter.toString(srcsfile);
	    DataInputStream dis = new DataInputStream(new StringBufferInputStream(buffer));
	    try {
		line = dis.readLine();
		while (line != null) {
		    SourceFileManager sfm = new SourceFileManager(this, line);
		    addChild(sfm);
		    while ((line = dis.readLine()) != null && line.charAt(0) == '\t') {
			ClassManager cm = new ClassManager(sfm, line.substring(1));
			sfm.addChild(cm);
		    }
		}
	    } finally {
		dis.close();
	    }
	} catch(java.io.FileNotFoundException e) {
	    SystemManager.errorln("Database file not found: " + srcsfile.getPath());
	} catch(java.io.IOException e) {
	    SystemManager.errorln("Database file cannot read: " + srcsfile.getPath());
	}
    }

    /**
     * create a package. it makes directory if it doesn't exist
     */
    boolean create() {
	TeikadeFile file = file();
	boolean exist = file.exists();
	if (!exist) {
	    file.mkdirs();
	}
	return rebuild(exist);
    }

    /**
     * rebuild database for package. force Transcript raised
     */
    public boolean rebuild() {
	return rebuild(true);
    }

    /**
     * rebuild database for package
     */
    public boolean rebuild(boolean raise) {
	SystemManager.message("Started rebuilding database for ");
	SystemManager.message(name());
	SystemManager.message(" package at ");
	SystemManager.messageln(new Date().toString());
	TeikadeFile dir = directory();
	if (raise) SystemManager.raiseTranscript();
	// thanks SCCS/RCS patch by Toshiki Murata.
	boolean result = rebuild(dir, StringSorter.sort(dir.list(new SourceFileFilter())));
	SystemManager.message("Finished rebuilding database for ");
	SystemManager.message(name());
	SystemManager.message(" package at ");
	SystemManager.messageln(new Date().toString());
	return result;
    }

    /**
     * rebuild database for package
     */
    public boolean rebuild(TeikadeFile dir, String files[]) {
	Thread currentThread = Thread.currentThread();
	int currentPriority = currentThread.getPriority();
	currentThread.setPriority(Thread.MIN_PRIORITY);
	try {
	    disposeChildren();
	    boolean needChanges = !((ProjectManager)parent).absolutePath().startsWith(TeikadeProperties.getString("java.home"));
	    for (int i = 0; i < files.length; i++) {
		if (files[i].endsWith(".java")||files[i].endsWith(".jav")||files[i].endsWith(".JAV")) {
		    Thread.yield();
		    System.gc();
		    try {
			SourceFileManager sfm = new SourceFileManager(this, files[i]);
			sfm.readSource();
			if (needChanges) {
			    Enumeration c = sfm.classes();
			    while (c.hasMoreElements()) {
				ClassManager cm = (ClassManager)c.nextElement();
				cm.rebuilded(cm.getCode());
				Enumeration m = cm.methods();
				while (m.hasMoreElements()) {
				    MethodManager mm = (MethodManager)m.nextElement();
				    mm.rebuilded(mm.getCode());
				}
			    }
			}
			addChild(sfm);
		    } catch (DescriptionException e) {
			SystemManager.errorln("Could not parse source file: " + e.getMessage());
		    }
		    catch (TeikadeException e) {
			SystemManager.errorln("Could not read source file: " + e.getMessage());
		    }
		}
	    }
	    updateSourcesFile();
	} finally {
	    currentThread.setPriority(currentPriority);
	}
	return child != null;
    }

    /**
     * rename a package
     */
    public Manager rename(String newName) {
	Manager ret = rename(newName, true);
	if (ret != null) {
	    ((ProjectManager)parent).updatePackagesFile();
	}
	return ret;
    }

    /**
     * rename a package
     */
    public Manager rename(String newName, boolean replacePackageStatements) {
	String newPath = pathFor(newName);
	TeikadeFile newFile = TeikadeFile.create(((PathedManager)parent).directory(), newPath);
	String oldName = name;
	TeikadeFile oldFile = file();
	TeikadeFile sourcesFile = sourcesFile();
	try {
	    newFile.mkdirs();
	    // thanks SCCS/RCS patch by Toshiki Murata
	    String files[] = oldFile.list(new SourceFileFilter());
	    boolean canRemove = true;
	    for (int i = 0; i < files.length; i++) {
		TeikadeFile f = TeikadeFile.create(oldFile, files[i]);
		if (f.isDirectory()) {
		    if (parent.findChild(oldName + "." + files[i]) != null) {
			// sub package is found
			canRemove = false;
		    } else if (files[i].equals(".") || files[i].equals("..")) {
			continue;
		    }
		} else if (files[i].endsWith(".java") && replacePackageStatements) {
		    SourceFileManager sfm = findSource(files[i]);
		    sfm.moveTo(oldName, newName);
		}
		f.renameTo(TeikadeFile.create(newFile, files[i]));
	    }
	    if (canRemove) {
		/*
		 old package directory should be removed. but
		 java.io.File has no method for remove file... sigh
		 oldFile.remove();
		 */
	    }
	    packages.remove(oldName);
	    packages.put(newName, this);
	    super.rename(newName);
	    TeikadeFile newSourcesFile = sourcesFile();
	    sourcesFile.renameTo(newSourcesFile);
	    setPath(newPath);
	    SystemManager.messageln("Package " + oldName + " is renamed to " + newName + ".");
	    return this;
	} catch (Exception e) {
	    SystemManager.exception("Could not rename package " + oldName + " to " + newName, e);
	    return null;
	}
    }

    public Enumeration sources() {
	return children();
    }

    public Enumeration sourceNames() {
	return childNames();
    }

    public int numSources() {
	return numChildren();
    }

    public Enumeration classes() {
	Vector classes = new Vector();
	for (Enumeration s = sources(); s.hasMoreElements(); ) {
	    SourceFileManager sfm = (SourceFileManager)s.nextElement();
	    for (Enumeration c = sfm.classes(); c.hasMoreElements(); ) {
		classes.addElement(c.nextElement());
	    }
	}
	return classes.elements();
	// return new ManagerEnumerator(child, 2);
    }

    public Enumeration classNames() {
	Vector classes = new Vector();
	for (Enumeration s = sources(); s.hasMoreElements(); ) {
	    SourceFileManager sfm = (SourceFileManager)s.nextElement();
	    for (Enumeration c = sfm.classes(); c.hasMoreElements(); ) {
		Manager m = (Manager)c.nextElement();
		classes.addElement(m.name());
	    }
	}
	return classes.elements();
	// return new ManagerEnumerator(child, 2);
    }

    public int numClasses() {
	int count = 0;
	for (Enumeration s = sources(); s.hasMoreElements(); ) {
	    SourceFileManager sfm = (SourceFileManager)s.nextElement();
	    count += sfm.numClasses();
	}
	return count;
    }

    public ClassManager findClass(String name) {
	for (Enumeration s = sources(); s.hasMoreElements(); ) {
	    SourceFileManager sfm = (SourceFileManager)s.nextElement();
	    ClassManager cm = sfm.findClass(name);
	    if (cm != null) {
		return cm;
	    }
	}
	return null;
    }

    public SourceFileManager findSourceForClass(String name) {
	for (Enumeration s = sources(); s.hasMoreElements(); ) {
	    SourceFileManager sfm = (SourceFileManager)s.nextElement();
	    if (sfm.findClass(name) != null) {
		return sfm;
	    }
	}
	return null;
    }

    public SourceFileManager findSource(String name) {
	return (SourceFileManager)findChild(name);
    }

    /**
     * create SourceFileManager and add it to the list
     * SFMs are sorted in dictionary order
     */
    public SourceFileManager createSourceFileManager(String fname) {
	SourceFileManager sfm = new SourceFileManager(this, fname);
	SourceFileManager prev = null;
	SourceFileManager next = null;
	Enumeration s = sources();
	while (s.hasMoreElements()) {
	    next = (SourceFileManager)s.nextElement();
	    if (next.name().compareTo(fname) > 0) break;
	    prev = next;
	    next = null;
	}
	if (prev == null) {
	    sfm.brother = child;
	    child = sfm;
	} else {
	    sfm.brother = next;
	    prev.brother = sfm;
	}
	sfm.setDefaultHeader();
	return sfm;
    }

    /**
     * create a new class from ClassSpec
     * this method is only called from Browser
     */
    public ClassManager createClass(ClassSpec spec) {
	try {
	    return createClass(spec, spec.description() + " {\n}\n");
	} catch (DescriptionException e) {
	    SystemManager.exception("ClassSpec produced illegal description", e);
	    return null;
	}
    }

    /**
     * create new class from ClassSpec and code
     * file name to add the class is requested to user
     */
    ClassManager createClass(ClassSpec spec, String newCode)
    throws DescriptionException {
	if (findClass(spec.className()) != null) {
	    SystemManager.beep();
	    SystemManager.information("Class " + spec.className() + " already exists.");
	    return null;
	}
	String fname = spec.className() + ".java";
	if (!spec.isPublic()) {
	    // public class description should go into className.java
	    fname = SystemManager.requestFileName(
			"What is the filename for the class " + spec.className() + "?",
			absolutePath(), fname);
	    // null indicates cancel create operation
	    if (fname == null) return null;
	}
	return createClass(spec, newCode, fname);
    }

    /**
     * internal class creation method
     * class name confliction should be checked before calling this method
     */
    private ClassManager createClass(ClassSpec spec, String newCode, String fname)
    throws DescriptionException {
	SourceFileManager sfm = findSource(fname);
	if (sfm == null) {
	    sfm = createSourceFileManager(fname);
	}
	if (!sfm.isEditable()) {
	    return null;
	}
	ClassManager cm = new ClassManager(sfm, spec);
	if (cm.resolveDependancy()) {
	    sfm.addChild(cm);
	    updateSourcesFile();
	    cm.created(newCode);
	    try {
		Enumeration m = cm.methods();
		while (m.hasMoreElements()) {
		    MethodManager mm = (MethodManager)m.nextElement();
		    mm.created(mm.getCode());
		}
	    } catch (TeikadeException e) {
		SystemManager.exception("Could not read source file: ", e);
	    }
	    return cm;
	} else {
	    return null;
	}
    }

    /**
     * returns package statment for this package
     */
    protected String packageStatement() {
	String name = name();
	if (name.indexOf(SEPARATOR_REPLACEMENT) >= 0) {
	    return "";
	} else {
	    return "package " + name + ";\n";
	}
    }

    /**
     * remove a source file from this package
     * update database
     */
    protected Manager removeChild(Manager child) {
	if (super.removeChild(child) != null) return child;
	updateSourcesFile();
	return null;
    }

    /**
     * dispose me
     */
    protected void dispose() {
	packages.remove(name);
	super.dispose();
    }
}
