/*
 * Copyright (c) 1995 PFU Limited.
 *	author Osamu Satoh
 */
package teikade.sys;

import teikade.util.ChangeList;
import teikade.util.StringSorter;
import teikade.util.CodeConverter;
import teikade.gui.Transcript;
import teikade.gui.TeikadeFrame;
import java.io.*;
import java.util.Vector;
import java.util.Hashtable;
import java.util.Enumeration;
import java.util.Date;

/**
 * project management class
 * children are instances of PackageManager
 */
public final class ProjectManager extends PathedManager {
    /**
     * the directory in which database for this project goes
     */
    private TeikadeFile profileDirectory;

    /**
     * change history for this project
     */
    private ChangeList changes;

    /**
     * default header file for this project
     */
    private TeikadeFile header = null;

    public Manager remove() {
	if (equals(SystemManager.getProjects())) {
	    SystemManager.setProjects((ProjectManager)brother);
	} else {
	    Enumeration elders = SystemManager.allProjects();
	    while (elders.hasMoreElements()) {
		ProjectManager prev = (ProjectManager)elders.nextElement();
		if (equals(prev.brother)) {
		    prev.brother = brother;
		    break;
		}
	    }
	}
	disposeChildren();
	// updatePackagesFile();
	SystemManager.updateProjects();
	return null;
    }

    ProjectManager(String name, String path) throws java.io.IOException {
	super(null, name, path);
	profileDirectory = TeikadeFile.create(path + File.separator + SystemManager.PROFDIRNAME);
	if (!profileDirectory.exists()) {
	    profileDirectory = SystemManager.profileDirectory();
	    changes = SystemManager.changes();
	} else {
	    changes = new ChangeList(TeikadeFile.create(profileDirectory, ChangeList.CHANGES));
	}
    }

    /**
     * get project CodePath
     */
    public CodePath getCodePath() {
	return new CodePath(null, this, null, null, null);
    }

    /**
     * whether it is editable or not
     */
    public boolean isEditable() {
	if (commentFile().exists()) {
	    return commentFile().canWrite();
	} else {
	    return directory().canWrite();
	}
    }

    /**
     * whether it has default header or not
     */
    public boolean hasDefaultHeader(boolean mustBeEditable) {
	if (headerFile().exists()) {
	    if (headerFile().canRead()) {
		return mustBeEditable ? headerFile().canWrite() : true;
	    }
	    SystemManager.information("Warning: The default header file for " + name +
					" exists but cannot read");
	}
	return false;
    }

    /**
     * set comment for this Project
     */
    public Manager setCode(String newComment) {
	try {
	    TeikadeFile commentFile = commentFile();
	    TeikadeFile backupFile = backupFile(commentFile);
	    TeikadeFile newFile = newFile(commentFile);
	    OutputStream os = newFile.getOutputStream();
	    try {
		os.write(CodeConverter.toBytes(newComment));
	    } finally {
		os.close();
	    }
	    commentFile.renameTo(backupFile);
	    newFile.renameTo(commentFile);
	    return this;
	} catch (java.io.IOException e) {
	    return null;
	}
    }

    /**
     * get project comment
     */
    protected String getCodeInternal() throws java.io.IOException {
	TeikadeFile commentFile = commentFile();
	if (commentFile.exists()) {
	    return CodeConverter.toString(commentFile);
	} else {
	    return "<body>Project comment for JavaDoc</body>";
	}
    }

    /**
     * get File object for project comment
     */
    TeikadeFile commentFile() {
	return TeikadeFile.create(directory(), "overview-" + name + ".html");
    }

    /**
     * get File object for project default header
     */
    TeikadeFile headerFile() {
	if (header == null) {
	    header = TeikadeFile.create(directory(), ".header");
	}
	return header;
    }

    /**
     * get ChangeList for this project
     */
    public ChangeList changes() {
	return changes;
    }

    /**
     * rebuild database for project
     */
    public boolean rebuild() {
	Thread currentThread = Thread.currentThread();
	int currentPriority = currentThread.getPriority();
	currentThread.setPriority(Thread.MIN_PRIORITY);
	SystemManager.raiseTranscript();
	SystemManager.message("Started rebuilding database for ");
	SystemManager.message(name());
	SystemManager.message(" project at ");
	SystemManager.messageln(new Date().toString());
	try {
	    disposeChildren();
	    Vector pkgs = listPackages(directory());
	    TeikadeFile packagesFile = packagesFile();
	    packagesFile.renameTo(backupFile(packagesFile));
	    if (pkgs.size() <= 0) return false;
	    PrintStream ps = new PrintStream(packagesFile.getOutputStream());
	    try {
		if (pkgs.size() > 0) {
		    for (Enumeration e = pkgs.elements(); e.hasMoreElements(); ) {
			currentThread.yield();
			PackageManager p = (PackageManager)e.nextElement();
			addChild(p);
			ps.println(p.name());
		    }
		    return true;
		} else {
		    return false;
		}
	    } finally {
		ps.close();
	    }
	} catch (java.io.IOException e1) {
	    SystemManager.exception("Could not write packages list on "
					+ packagesFile().getPath() + ".", e1);
	    return false;
	} finally {
	    SystemManager.message("Finished rebuilding database for ");
	    SystemManager.message(name());
	    SystemManager.message(" project at ");
	    SystemManager.messageln(new Date().toString());
	    currentThread.setPriority(currentPriority);
	}
    }

    /**
     * list up packages for project
     * it is called recursive
     * also lists classes for packages
     * @return package list
     */
    public Vector listPackages(TeikadeFile dir) throws java.io.IOException {
	PackageManager pkgMan = null;
	Vector sub = new Vector();
	// thanks SCCS/RCS patch by Toshiki Murata
	String files[] = StringSorter.sort(dir.list(new SourceFileFilter()));
	if (files == null) {
	    // no files and directories exist
	    return sub;
	}
	for (int i = 0; i < files.length; i++) {
	    String fn = files[i];
	    TeikadeFile f = TeikadeFile.create(dir, fn);
	    if (f.isDirectory()) {
		Vector v = listPackages(f);
		for (Enumeration e = v.elements(); e.hasMoreElements(); ) {
		    sub.addElement(e.nextElement());
		}
	    } else if (fn.endsWith(".java")||fn.endsWith(".jav")||fn.endsWith(".JAV")) {
		if (pkgMan == null) {
		    if (dir == directory()) {
			// is top directory
			pkgMan = new TopDirPackageManager(this);
		    } else {
			String path = dir.getPath().substring(path().length() + 1);
			pkgMan = new PackageManager(this, PackageManager.nameFor(path), path);
		    }
		}
	    }
	}
	if (pkgMan != null) {
	    sub.insertElementAt(pkgMan, 0);
	    pkgMan.rebuild(dir, files);
	}
	return sub;
    }

    /**
     * add package entry
     */
    public PackageManager createPackageAfter(String pkgname, PackageManager pm) {
	if (PackageManager.findPackage(pkgname) != null) {
	    SystemManager.information("Package already exists: " + pkgname, "Abort");
	    return null;
	} else if (!directory().canWrite()) {
	    SystemManager.information("Project directory isn't writable:\n" +
					directory().getPath().toString(), "Abort");
	    return null;
	}
	PackageManager pman = PackageManager.forName(this, pkgname, false);
	if (pman != null) {
	    if (pm == null) {
		addChild(pman);
	    } else {
		pman.brother = pm.brother;
		pm.brother = pman;
	    }
	    pman.create();
	    updatePackagesFile();
	    return pman;
	}
	return null;
    }

    public TeikadeFile profileDirectory() {
	return profileDirectory;
    }

    public int numPackages() {
	return numChildren();
    }

    public Enumeration packages() {
	return children();
    }

    public Enumeration packageNames() {
	return childNames();
    }

    public PackageManager findPackage(String name) {
	return (PackageManager)findChild(name);
    }

    /**
     * get database file contains package list of this Project
     */
    TeikadeFile packagesFile() {
	return TeikadeFile.create(profileDirectory, name);
    }

    /**
     * read the file contains package list in this project
     */
    void readPackagesFile() {
	if (child != null) {
	    disposeChildren();
	}
	TeikadeFile pkgsfile = packagesFile();
	try {
	    if (!pkgsfile.exists()) return;
	    String line;
	    String pr = null;
	    DataInputStream dis = new DataInputStream(pkgsfile.getInputStream());
	    try {
		while ((line = dis.readLine()) != null) {
		    PackageManager pman = PackageManager.forName(this, line);
		    addChild(pman);
		    pman.readSourcesFile();
		}
	    } finally {
		dis.close();
	    }
	} catch(java.io.IOException e) {
	    SystemManager.exception("Could not read package list file: " + pkgsfile.getPath(), e);
	}
    }

    /**
     * update package list database file
     */
    void updatePackagesFile() {
	TeikadeFile pkgsFile = packagesFile();
	try {
	    String line;
	    String pr = null;
	    TeikadeFile backupFile = backupFile(pkgsFile);
	    TeikadeFile newFile = newFile(pkgsFile);
	    PrintStream ps = new PrintStream(newFile.getOutputStream());
	    try {
		Enumeration p = packages();
		while (p.hasMoreElements()) {
		    PackageManager pman = (PackageManager)p.nextElement();
		    ps.println(pman.name());
		}
	    } finally {
		ps.close();
	    }
	    pkgsFile.renameTo(backupFile);
	    newFile.renameTo(pkgsFile);
	} catch(java.io.IOException e) {
	    SystemManager.exception("Could not update package list file: " + pkgsFile.getPath(), e);
	}
    }

    /**
     * remove a package from this project
     */
    protected Manager removeChild(Manager child) {
	if (super.removeChild(child) != null) return child;
	updatePackagesFile();
	return null;
    }

    /**
     * rename this project
     */
    public Manager rename() throws TeikadeException {
	String newName =
	    SystemManager.request("new name for " + name, name);
	if (newName == null || newName.equals(name)) return null;
	if (SystemManager.findProject(newName) != null) {
	    SystemManager.information(
		"Specified name is already used for another project");
	    return null;
	}
	TeikadeFile dir = profileDirectory();
	String files[] = dir.list();
	for (int i = 0; i < files.length; i++) {
	    String newFname = null;
	    if (files[i].equals(name)) {
		newFname = newName;
	    } else if (files[i].startsWith(name + ".")) {
		newFname = newName + files[i].substring(name.length());
	    } else continue;
	    TeikadeFile from = TeikadeFile.create(dir, files[i]);
	    TeikadeFile to = TeikadeFile.create(dir, newFname);
	    from.renameTo(to);
	}
	Transcript.getTranscriptMenuBar().renameProject(name, newName);
	Manager retm = super.rename(newName);
	SystemManager.updateProjects();
	return retm;
    }
}
