/*
 * Copyright (c) 1998 PFU Limited.
 *	author Kazuhisa Shirakami
 */
package teikade.util;

public class TabConverter {
    private static final boolean DEBUG = false;
    private static final boolean JDK102 = true;

    public static String untabify(String text, int tabWidth) {
	TabbedSource src = new TabbedSource(text, tabWidth);
	TabbedTextBuffer dst = new TabbedTextBuffer(tabWidth);
	char c;
	while (src.hasMoreChar()) {
	    c = src.nextChar();
	    if (src.inQuote()) {
		dst.append(c);
	    } else if (c == '\t') {
		dst.fillSpace(src.column);
	    } else {
		dst.append(c);
	    }
	}
	return dst.toString();
    }

    public static String convert(String text, int textTab, int convertTab) {
	TabbedSource src = new TabbedSource(text, textTab);
	TabbedTextBuffer dst = new TabbedTextBuffer(convertTab);
	char c = '\0', lastChar;
	int lastSrcColumn;
	boolean inQuote;
	while (src.hasMoreChar()) {
	    lastChar = c;
	    lastSrcColumn = src.column;
	    inQuote = src.inQuote();
	    c = src.nextChar();
	    if (DEBUG) {
		    System.out.println(c + ":" + src.column()
					+ " Q:" + src.inQuote()
					+ " C:" + src.inComment());
	    }
	    if (inQuote) {
		dst.append(c);
	    } else if (c != ' ' && c != '\t') {
		if (dst.column < lastSrcColumn) {
		    if (dst.column == 0 || lastChar == '\t'
		     || ((lastSrcColumn % convertTab) == 0
			 && dst.column + 1 < lastSrcColumn)) {
			if (DEBUG) {
			    System.out.println("fillTab:" + lastSrcColumn);
			}
			dst.fillTab(lastSrcColumn);
		    } else {
			if (DEBUG) {
			    System.out.println("fillSpace:" + lastSrcColumn);
			}
			dst.fillSpace(lastSrcColumn);
		    }
		}
		if (DEBUG) {
		    System.out.println("append:" + c);
		}
		dst.append(c);
	    }
	}
	if (DEBUG) {
	    return "";
	} else {
	    return dst.toString();
	}
    }

    public static int index(String text, int index, int fromTab, int toTab) {
	TabbedText from = new TabbedText(text, toTab);
	for (int i = index; --i >= 0 && from.hasMoreChar(); ) {
	    from.nextChar();
	}
	int column = from.column();
	TabbedText to = new TabbedText(convert(text, toTab, fromTab), fromTab);
	while (to.hasMoreChar() && to.column < column) {
	    to.nextChar();
	}
	return to.index();
    }

    public static void main(String args[]) throws Exception {
	int inTab = Integer.parseInt(args[0]);
	int outTab = Integer.parseInt(args[1]);
	java.io.DataInputStream in = new java.io.DataInputStream(System.in);
	String line;
	while ((line = in.readLine()) != null) {
	    System.out.println(convert(line, inTab, outTab));
	}
    }
}

class ColumnManager {
    protected int tabWidth;
    protected int column;

    protected ColumnManager(int tabWidth) {
	this.tabWidth = tabWidth;
    }

    protected char add(char c) {
	if (c == '\t') {
	    column += tabWidth - (column % tabWidth);
	} else if (c == '\n' || c == '\r') {
	    column = 0;
	} else if (isDoubleWidth(c)) {
	    column += 2;
	} else {
	    column++;
	}
	return c;
    }

    public int column() {
	return column;
    }

    public static boolean isDoubleWidth(char c) {
	return TextUtil.getInstance().isDoubleWidth(c);
    }
}

class TabbedText extends ColumnManager {
    String text;
    int length;
    int current = 0;

    public TabbedText(String text, int tabWidth) {
	super(tabWidth);
	this.text = text;
	this.length = text.length();
    }

    public boolean hasMoreChar() {
	return current < length;
    }

    public char nextChar() {
	return add(text.charAt(current++));
    }

    public int index() {
	return current;
    }
}

class TabbedTextBuffer extends ColumnManager {
    private StringBuffer buffer = new StringBuffer();

    public TabbedTextBuffer(int tabWidth) {
	super(tabWidth);
    }

    public void append(char c) {
	buffer.append(add(c));
    }

    public void fillSpace(int toColumn) {
	int n = toColumn - column;
	while (--n >= 0) {
	    append(' ');
	}
    }

    public void fillTab(int toColumn) {
	int to = toColumn - (toColumn % tabWidth);
	while (column < to) {
	    append('\t');
	}
	if (to < toColumn) {
	    fillSpace(toColumn);
	}
    }

    public String toString() {
	return buffer.toString();
    }
}

class TabbedSource extends TabbedText {
    boolean multiComment = false;
    boolean singleComment = false;
    char quote = '\0';
    boolean escape;

    public TabbedSource(String source, int tabWidth) {
	super(source, tabWidth);
    }

    public char nextChar() {
	char c = super.nextChar();
	if (quote != '\0') {
	    if (escape) {
		escape = false;
	    } else if (c == quote) {
		quote = '\0';
	    } else if (c == '\\') {
		escape = true;
	    }
	} else if (multiComment) {
	    if (escape) {		// founded '*'
		escape = false;
		if (c == '/') {
		    multiComment = false;
		if (c == '*') {
		    escape = true;
		}
	    }
	    } else if (c == '*') {
		escape = true;
	    }
	} else if (singleComment) {
	    if (c == '\n' || c == '\r') {
		singleComment = false;
	    }
	} else {
	    if (escape) {		// founded '/'
		escape = false;
		if (c == '/') {
		    singleComment = true;
		    return c;
		}
		if (c == '*') {
		    multiComment = true;
		    return c;
		}
	    }
	    if (c == '/') {
		escape = true;
	    } else if (c == '"' || c == '\'') {
		quote = c;
	    }
	}
	return c;
    }

    public boolean inComment() {
	return multiComment || singleComment;
    }

    public boolean inQuote() {
	return quote != '\0';
    }
}
